; Program for VAmeterUSB
; Version: 0.0
; Last updated:
; History of versions:
;	1.0 First released version

#include p16f818.inc   ;include file with config bit definitions
	__config _CP_OFF & _HS_OSC & _WDT_ON & _LVP_OFF  & _MCLR_ON ;Set oscillator, WDT, MCLR etc.

;*****************   Communication Parameters   *********************
; The folloving delay constants are for 19200 Baud
; 6 MHz Clock input (CLKOUT = Clock input / 4)
;
BAUD_1	equ	.24		; 6+3X = CLKOUT/Baud
BAUD_X	equ	.20		; 17+3X = CLKOUT/Baud
;
DX	equ	2		; Transmit Pin (Bit 2 of Port A)
;
;********************************************************************
; My private instructions
#DEFINE	bank0	bcf STATUS,RP0
#DEFINE	bank1	bsf STATUS,RP0
;
; Output port configuration
#define INP10     bcf	PORTB,.2  ; Attenuator setting
#define INP100    bsf	PORTB,.2
#define GAIN3_ON  bcf	PORTB,.1  ; Amp 3x setting
#define GAIN3_OFF bsf	PORTB,.1
#define GAIN8_ON  bcf	PORTB,.3  ; Amp 8x setting
#define GAIN8_OFF bsf	PORTB,.3
#define LED_ON    bcf	PORTA,.4  ; LED controll
#define LED_OFF   bsf	PORTA,.4
;****************  Data RAM Assignments  ***************************
;
	ORG	20H		; Dummy Origin
XmtReg	RES	1		; Data to be transmitted
Count	RES	1		; Counter for #of Bits Transmitted
DlyCnt	RES	1		; Delay counter for serial interface
DelMils	RES	1		; Milisecond counter
DelSecs	RES	1		; Second counter

VRange	RES	1		; Voltage Range
IRange	RES	1		; Current Range
ZeroLow	RES	1		; Zero voltage value
ZeroHig	RES	1
MaxLow	RES	1		; Maximum value
MaxHig	RES	1
Ipoint	RES	1		; Auxiliary pointer "i"
AuxLow	RES	1		; Auxiliary value
AuxHig	RES	1
RangeFl	RES	1		; Set Range/Otput Data flag

; Arithmetics registers
L_byte	RES	1
H_byte	RES	1
temp	RES	1
R0	RES	1
R1	RES	1
R2	RES	1
;
	ORG	40H		; Dummy Origin
ADVal	RES	1		; Here must start data array!!!
;*******************************************************************
	ORG	0
; Clear output port Pins
	bsf	PORTA,DX
	bcf	PORTA,.4
; I/O port configuration
	bank1
	movlw	0xAB	; Pins of port A
	movwf	TRISA
	movlw	0x81	; Pins of port B
	movwf	TRISB
; Configuration of A/D converter
	movlw	0xC4	;Right justify, AN0,AN1,AN3 - analog, AN2 - digital
	movwf	ADCON1
	bank0	
	movlw	0x59	;Tad oscil., Channel
	movwf	ADCON0
; Variable defaut values
	movlw	.7
	movwf	VRange
	movlw	.3
	movwf	IRange
	clrf	ZeroLow
	movlw	.2
	movwf	ZeroHig
; ****************************************
; Main program loop
; ****************************************
Main	clrwdt
	movlw	0x8A
; Send VRange character
	call	Xmtr
	movlw	0x41
	addwf	VRange,W
	call	Xmtr
; Switch 3x and 8x Amplifiers ON/OFF
	call	ADMUXGND
	movf	VRange,W
	andlw	0x07		; Suppose error value

	addwf	PCL,F		; Switch
	goto	AllON		; VRange = 0
	goto	G8ON		; VRange = 1
	goto	G3ON		; VRange = 2
	goto	AllON		; VRange = 3
	goto	AllOFF		; VRange = 4
	goto	G8ON		; VRange = 5
	goto	G3ON		; VRange = 6
AllOFF	GAIN3_OFF			; VRange = 7
	GAIN8_OFF
	goto	ComGND
G3ON	GAIN3_ON
	GAIN8_OFF
	goto	ComGND
G8ON	GAIN8_ON
	GAIN3_OFF
	goto	ComGND
AllON	GAIN3_ON
	GAIN8_ON
; Send GND measured values
ComGND	movlw	0x9A;
	call	Send32

; Send REFERENCE measured values
	movf	VRange,W
	andlw	0x07		; Suppose error value

	addwf	PCL,F		; Switch
	goto	Ref02		; VRange = 0
	goto	Ref12		; VRange = 1
	goto	Ref05		; VRange = 2
	goto	Ref02		; VRange = 3
	goto	Ref25		; VRange = 4
	goto	Ref12		; VRange = 5
	goto	Ref05		; VRange = 6
Ref25	call	ADMUXR25		; VRange = 7
	goto	ComMeas
Ref02	call	ADMUXR02
	goto	ComMeas
Ref05	call	ADMUXR05
	goto	ComMeas
Ref12	call	ADMUXR12
; Average GND value
ComMeas	call	Averag
	movlw	0x8D
; Not neccessary to meas REF in blind cycle
	movf	RangeFl,F
	skpnz
	call	Send32

; Measure Value - switch input atenuator
	movf	VRange,W
	andlw	0x07		; Suppose error value

	addwf	PCL,F		; Switch
	goto	LowInp		; VRange = 0
	goto	LowInp		; VRange = 1
	goto	LowInp		; VRange = 2
	goto	HigInp		; VRange = 3
	goto	LowInp		; VRange = 4
	goto	HigInp		; VRange = 5
	nop			; VRange = 6
HigInp	call	ADMUX100		; VRange = 7
	INP100
	goto	ComMax
LowInp	INP10
	call	ADMUX10
; Send measured values and find maximum
ComMax	movlw	0x9D;
	call	Send32

	call	GetMax
; Test new VRange for next measure
	movlw	0x01		; LH_byte = 340
	movwf	H_byte
	movlw	0x54
	movwf	L_byte
	movf	MaxLow,W
	movwf	R0
	movf	MaxHig,W
	movwf	R1
	call	Odecti
; if((Max>340) && (VRange<7)) VRange++;
	btfss	H_byte,.7
	goto	NoUpVolt
	movlw	0xF9
	addwf	VRange,W
	skpc
	incf	VRange,F
	goto	EndVolt
NoUpVolt	clrf	H_byte		; LH_byte = 100
	movlw	0x64
	movwf	L_byte
	movf	AuxLow,W
	movwf	R0
	movf	AuxHig,W
	movwf	R1
	call	Odecti
;	if((Max<100) && (VRange>0)) VRange--;
	btfsc	H_byte,.7
	goto	EndVolt
	movf	VRange,F
	skpz
	decf	VRange,F
EndVolt

; *** Current part ***
; Send IRange character
	movlw	0x8E
	btfsc	PORTB,.0
	movlw	0x8F
	call	Xmtr
	movlw	0x41
	addwf	IRange,W
	call	Xmtr
; Switch 3x and 8x Amplifiers ON/OFF
	call	ADMUXGND
	movf	IRange,W
	andlw	0x03		; Suppose error value

	addwf	PCL,F		; Switch
	goto	AllONI		; IRange = 0
	goto	G8ONI		; IRange = 1
	goto	G3ONI		; IRange = 2
	GAIN3_OFF			; IRange = 3
	GAIN8_OFF
	goto	ComGNDI
G3ONI	GAIN3_ON
	GAIN8_OFF
	goto	ComGNDI
G8ONI	GAIN8_ON
	GAIN3_OFF
	goto	ComGNDI
AllONI	GAIN3_ON
	GAIN8_ON
; Send GND measured values
ComGNDI	movlw	0x9E
	call	Send32

; Send REFERENCE measured values
	movf	IRange,W
	andlw	0x03		; Suppose error value

	addwf	PCL,F		; Switch
	goto	Ref02I		; IRange = 0
	goto	Ref12I		; IRange = 1
	goto	Ref05I		; IRange = 2
	call	ADMUXR25		; IRange = 3
	goto	ComMeasI
Ref02I	call	ADMUXR02
	goto	ComMeasI
Ref05I	call	ADMUXR05
	goto	ComMeasI
Ref12I	call	ADMUXR12
; Average GND value
ComMeasI	call	Averag
	movlw	0xF8
	call	Send32

; Measure Value - switch input atenuator
	call	ADMUXCUR
	movlw	0xD8
	call	Send32
	call	GetMax
; Test new IRange for next measure
	movlw	0x01		; LH_byte = 340
	movwf	H_byte
	movlw	0x54
	movwf	L_byte
	movf	MaxLow,W
	movwf	R0
	movf	MaxHig,W
	movwf	R1
	call	Odecti
; if((Max>340) && (IRange<3)) IRange++;
	btfss	H_byte,.7
	goto	NoUpCurr
	movlw	0xFD
	addwf	IRange,W
	skpc
	incf	IRange,F
	goto	EndCurr
NoUpCurr	clrf	H_byte		; LH_byte = 100
	movlw	0x64
	movwf	L_byte
	movf	AuxLow,W
	movwf	R0
	movf	AuxHig,W
	movwf	R1
	call	Odecti
;	if((Max<100) && (IRange>0)) IRange--;
	btfsc	H_byte,.7
	goto	EndCurr
	movf	IRange,F
	skpz
	decf	IRange,F
; If Odd period, just new range was set
EndCurr	incf	RangeFl,F
	movlw	0x03
	andwf	RangeFl,F
	skpz
	goto	Main
; At last wait for next measure - blink LED
	LED_ON
	call	Del01sec
	LED_OFF
	goto	Main
;
;*****************************************
;                 Subroutines
;*****************************************
; Get maximum value in array of values to Max (copy is in Aux as well)
GetMax	clrf	MaxLow
	clrf	MaxHig
	movlw	0x40
	movwf	FSR
	movlw	.32
	movwf	Ipoint
LopMax	movf	ZeroLow,W
	movwf	R0
	movf	ZeroHig,W
	movwf	R1
	movf	INDF,W		; Prepare for Value - Zero
	movwf	L_byte
	incf	FSR,F
	movf	INDF,W
	movwf	H_byte
	incf	FSR,F
	call	Odecti		; Value - Zero
	btfsc	H_byte,.7
	call	Minus		; Absolute value
	movf	L_byte,W
	movwf	AuxLow
	movf	H_byte,W
	movwf	AuxHig		; Save for future max
	movf	MaxLow,W
	movwf	R0
	movf	MaxHig,W
	movwf	R1
	call	Odecti		; |Value| - Max
	btfsc	H_byte,.7
	goto	NoNewMax		; Set new maximum
	movf	AuxLow,W
	movwf	MaxLow
	movf	AuxHig,W
	movwf	MaxHig
NoNewMax	decfsz	Ipoint,F
	goto	LopMax
	movf	MaxLow,W
	movwf	AuxLow
	movf	MaxHig,W
	movwf	AuxHig
	return
; Store in Zero average value from array
Averag	clrf	L_byte
	clrf	H_byte
	movlw	0x40
	movwf	FSR
	movlw	.32
	movwf	Ipoint
LoopAv	movf	INDF,W		; Sum all values
	movwf	R0
	incf	FSR,F
	movf	INDF,W
	movwf	R1
	incf	FSR,F
	call	Secti
	decfsz	Ipoint,F
	goto	LoopAv
	movlw	.32		; Get average
	movwf	R0
	call	DivMod
	movf	L_byte,W		; Store it in Zero
	movwf	ZeroLow
	movf	H_byte,W
	movwf	ZeroHig
	return

; Multiplexer address setting
ADMUX10	bcf	PORTB,.4
	bcf	PORTB,.5
	bcf	PORTB,.6
	return  
ADMUX100	bsf	PORTB,.4
	bcf	PORTB,.5
	bcf	PORTB,.6
	return  
ADMUXCUR	bcf	PORTB,.4
	bsf	PORTB,.5
	bcf	PORTB,.6
	return  
ADMUXGND	bsf	PORTB,.4
	bsf	PORTB,.5
	bcf	PORTB,.6
	return  
ADMUXR25	bcf	PORTB,.4
	bcf	PORTB,.5
	bsf	PORTB,.6
	return  
ADMUXR02	bsf	PORTB,.4
	bcf	PORTB,.5
	bsf	PORTB,.6
	return  
ADMUXR05	bcf	PORTB,.4
	bsf	PORTB,.5
	bsf	PORTB,.6
	return  
ADMUXR12	bsf	PORTB,.4
	bsf	PORTB,.5
	bsf	PORTB,.6
	return  
	
; Transmitter
;
Xmtr	movf	RangeFl,F	; Output data only in measure period
	skpz
	return
	movwf	XmtReg
	movlw	.8
	movwf	Count
;
	bcf	PORTA,DX	; Send Start Bit
	call	Delay1
X_next	bcf	STATUS,C
	rrf	XmtReg,F	; LSB first
	btfsc	STATUS,C
	bsf	PORTA,DX
	btfss	STATUS,C
	bcf	PORTA,DX
	call	DelayX
	decfsz	Count,F
	goto	X_next
	bsf	PORTA,DX	; Send Stop Bit
	call	Delay1
	call	Delay1
	call	Delay1
;
	retlw	.0
;	End of Transmission
;
DelayX	nop
	movlw	BAUD_X
	goto	save
Delay1	movlw	BAUD_1		; 52 usec for 19200 Baud
save	movwf	DlyCnt
redo_1	decfsz	DlyCnt,F
	goto	redo_1
	return
;
; 4 milisec delay
Del4ms	movlw	.24
	movwf	DelMils
Lop1ms	movlw	.81
	movwf	DlyCnt
	call	redo_1
	decfsz	DelMils,F
	goto	Lop1ms
	return
; 100 msec delay
Del01sec	movlw	.25
DelN4ms	movwf	DelSecs
Lop1s	call	Del4ms
	decfsz	DelSecs,F
	goto	Lop1s
	return

; AD convert (without channel selection)
; Takes 124 usec including "call"
ADCon	movlw	.42
	movwf	DlyCnt
Sample	decfsz	DlyCnt,F
	goto	Sample
	bsf	ADCON0,GO_DONE	; Start ...
ADLoop	btfsc	ADCON0,GO_DONE
	goto	ADLoop		; to stop takes 32 usec
	bank1
	movf	ADRESL,W
	bank0
	movwf	L_byte
	movf	ADRESH,W
	movwf	H_byte
	return			; and final 8 usec

; Send 32 integer values (2x 6bits, LOWBYTE first)
Send32	call	Xmtr	; Send byte from W at first
	call	Del4ms	; Wait for input stabilization
	movlw	0x40
	movwf	FSR
	movlw	.32
	movwf	Ipoint
Loop32A	call	ADCon	; Measure 32 sample values - 124 usec
	movf	L_byte,W
	movwf	INDF
	incf	FSR,F
	movf	H_byte,W
	movwf	INDF
	incf	FSR,F	; + 4 usec
	movlw	.245
	call	save	; + 4 + 2*W
	decfsz	Ipoint,F
	goto	Loop32A	; + 2 usec = total 625 usec
; Send data to output
	movf	RangeFl,F	; Output data only in measure period
	skpz
	return
	movlw	0x40
	movwf	FSR
	movlw	.32
	movwf	Ipoint
Loop32W	movf	INDF,W
	movwf	R0
	incf	FSR,F
	movf	INDF,W
	movwf	R1
	incf	FSR,F

	movlw	0x3F	; First LOW byte
	andwf	R0,W
	iorlw	0x40
	call	Xmtr
	rlf	R0,F	; Then HIGH byte
	rlf	R1,F
	rlf	R0,F
	rlf	R1,W
	andlw	0x0F
	iorlw	0x40
	call	Xmtr

	decfsz	Ipoint,F
	goto	Loop32W
	return

; -----------------------------------------------------------
; First operand in L_byte, H_byte, second in R0,R1 - result in L_byte, H_byte
; Residue e.g. in temp

; Change sign of L/H
Minus	comf	L_byte,F
	comf	H_byte,F
	incf	L_byte,F
	skpnz
	incf	H_byte,F
	return
; Subtract R0,R1 from L/H_byte
Odecti	comf	R0,F
	comf	R1,F
	incf	R0,F
	skpnz
	incf	R1,F
; v v v continue!!!! v v v
; Add L/H_and R0,R1
Secti	movf	R0,W
	addwf	L_byte,F
	skpnc
	incf	H_byte,F
	movf	R1,W
	addwf	H_byte,F
	return

; Divide L/H_byte by R0 (max 255), 
; Result in L/H_byte (copy of result in R1, R2), residue in temp
; Used: R1,R2,count and temp
DivMod	movlw	.16
	movwf	Count
; clear result and residue
	clrf	R1
	clrf	R2
	clrf	temp
LoDiMo	clrc
	rlf	L_byte,F
	rlf	H_byte,F
	rlf	temp,F
	skpnc
	goto	DivOver
	clrc
	rlf	R1,F
	rlf	R2,F
; test, is residue greater than divisor?
	comf	temp,W
	addwf	R0,W
	skpnc
	goto	NeZbVe
JeVeZb	incf	R1,F
	skpnz
	incf	R2,F
	movf	R0,W
	subwf	temp,F
; Residue is not over divisor
NeZbVe	decfsz	Count,F
	goto	LoDiMo
	movf	R1,W
	movwf	L_byte
	movf	R2,W
	movwf	H_byte
	return
; Residue overflow 255
DivOver	clrc
	rlf	R1,F
	rlf	R2,F
	goto	JeVeZb

	END
