/* ========================================================================== */
/*   Vmetr.c                                                                  */
/*   (c) 2007 St                                                              */
/*                                                                            */
/*   Program for Vmetr controlled ATtiny26                                    */
/*   Xtall frequency: 6 MHz                                                   */
/* ========================================================================== */

// Warning! Don't use bit manipulation suggested in WinAVR examples!
// Setting bit e.g. PORTB |= _BV(PB7) this way can cause problem!
// Value from port is read at first, modificated and written at last.
// Interrupt (just right after read port value) working with the same 
// port can modificate port value. Following write will destroy this data
// written by ISR (interrupt-service-routine). Use direct bit manipulation
// instead (asm instructions "sbi" and "cbi").

#include <avr/io.h>

// Constant definition for 19200 Bd
#define TXSPEED 28

// Definition of PIN connection PORTA
#define INP10     asm ("cbi 0x1B,3" ::)  // Attenuator setting
#define INP100    asm ("sbi 0x1B,3" ::)
#define GAIN3_ON  asm ("cbi 0x1B,5" ::)  // Amp 3x setting
#define GAIN3_OFF asm ("sbi 0x1B,5" ::)
#define GAIN8_ON  asm ("cbi 0x1B,2" ::)  // Amp 8x setting
#define GAIN8_OFF asm ("sbi 0x1B,2" ::)
#define LED_ON    asm ("cbi 0x1B,6" ::)  // LED controll
#define LED_OFF   asm ("sbi 0x1B,6" ::)
// MUX address setting
#define ADMUX10  asm ("cbi 0x18,1" ::);asm ("cbi 0x18,2" ::);asm ("cbi 0x18,3" ::)  
#define ADMUX100 asm ("sbi 0x18,1" ::);asm ("cbi 0x18,2" ::);asm ("cbi 0x18,3" ::)  
#define ADMUXCUR asm ("cbi 0x18,1" ::);asm ("sbi 0x18,2" ::);asm ("cbi 0x18,3" ::)  
#define ADMUXGND asm ("sbi 0x18,1" ::);asm ("sbi 0x18,2" ::);asm ("cbi 0x18,3" ::)  
#define ADMUXR25 asm ("cbi 0x18,1" ::);asm ("cbi 0x18,2" ::);asm ("sbi 0x18,3" ::)  
#define ADMUXR02 asm ("sbi 0x18,1" ::);asm ("cbi 0x18,2" ::);asm ("sbi 0x18,3" ::)  
#define ADMUXR05 asm ("cbi 0x18,1" ::);asm ("sbi 0x18,2" ::);asm ("sbi 0x18,3" ::)  
#define ADMUXR12 asm ("sbi 0x18,1" ::);asm ("sbi 0x18,2" ::);asm ("sbi 0x18,3" ::)  
// Definition of PIN connection PORTB
#define TX_ON    asm ("cbi 0x1B,7" ::)  // Serial TX setting
#define TX_OFF   asm ("sbi 0x1B,7" ::)

// Array for measured values
unsigned int Values[32];
unsigned char TXchar,VRange=7,IRange=3;
unsigned int ZeroVolt=512;
int Max;
int Zero,AuxVal;
unsigned char i,RangeFl=1;

// Send 1 character to serial line
void Sendchar(void) {
  unsigned char i,j;
  
  if(RangeFl) return;
  // Start bit
  TX_ON;
  for(j=0; j<TXSPEED; j++) {}
  // Send 8 bits, LSB first
  for(i=0; i<8; i++) {
    if(TXchar & 0x01) TX_OFF; else TX_ON;
    for(j=0; j<TXSPEED; j++) {}
	TXchar = TXchar >> 1;
  } 
  // Stop bit
  TX_OFF;
  for(j=0; j<TXSPEED; j++) {}
}


// Send 32 integer values (2x 6bits, LOWBYTE first)
void Send32Int()
{
  unsigned int cekej;
  unsigned char i,j;

  // Send 1 character (type of value)
  Sendchar();
  // Wait for input stabilization
  for(cekej=0; cekej<1000; cekej++) {}
  // Measure 32 sample values
  for(i=0; i<32; i++) {
	ADCSR |= 0x40;
	while (ADCSR & 0x40) {}
	Values[i] = ADC;
	for(j=0; j<200; j++) {}		// Wait for next sample
	for(j=0; j<124; j++) {}
  }
  if(RangeFl) return;
  // Send data to output
  for(i=0; i<32; i++) {
	TXchar = Values[i] & 0x3F;
	TXchar |= 0x40;
	Sendchar();
	TXchar = Values[i] >> 6;
	TXchar |= 0x40;
	Sendchar();
  }
}

// Find maximum value in array (with measuring before it)
void GetMax()
{
	Send32Int();
	Max=0;
	for(i=0;i<32;i++) {
		AuxVal=Values[i]-Zero;
		if(AuxVal<0) AuxVal=-AuxVal;
		if(Max<AuxVal) Max=AuxVal;
	}
}

// ****************************
//    Main program function  //
// ****************************
int main(void) {
  /* Enable PORTB as output (except PB6,5,4)*/
  DDRB = 0x8F;
  /* Enable PORTA as output (except PA2 - analog input)*/
  DDRA = 0xFD;
  /* Disable pull-ups*/
  MCUCR |= 0x40;
  // Set TX line to 1
  TX_OFF;

  // ADC setting
  ADMUX = 0x01;
  ADCSR = 0xC5;

  /***** Main loop of program *****/
  for(;;) {
    // *** Voltage part ***
	// Send VRange character
	TXchar = 0x8A;
	Sendchar();
	TXchar = VRange + 0x41;
	Sendchar();
	// Switch 3x and 8x Amplifiers ON/OFF
	ADMUXGND;
	switch(VRange) {
	case 0: 
	case 3: GAIN3_ON; GAIN8_ON; break;
	case 1:
	case 5: GAIN3_OFF; GAIN8_ON; break;
	case 2:
	case 6: GAIN3_ON; GAIN8_OFF; break;
	case 4:
	case 7: GAIN3_OFF; GAIN8_OFF; break;
	}
	// Send GND measured values
	TXchar = 0x9A;
	Send32Int();

	// Send REFERENCE measured values
	switch(VRange) {
	case 0:
	case 3: ADMUXR02; break;
	case 1:
	case 5: ADMUXR12; break;
	case 2:
	case 6: ADMUXR05; break;
	case 4:
	case 7: ADMUXR25; break;
	}
	// Average GND value
	Zero = 0;
	for(i=0;i<32;i++) Zero+=Values[i];
	Zero = Zero >> 5;
	TXchar = 0x8D;
	if(!RangeFl) Send32Int();

	// Measure Value - switch input atenuator
	switch(VRange) {
	case 0:
	case 1:
	case 2:
	case 4: {INP10;ADMUX10;} break;
	case 3:
	case 5:
	case 6:
	case 7: {INP100;ADMUX100;} break;
	}
	// Send measured values and find maximum
	TXchar = 0x9D;
	GetMax();

	// Test new VRange for next measure
	if((Max>340) && (VRange<7)) VRange++;
	if((Max<100) && (VRange>0)) VRange--;

    // *** Current part ***
	// Send IRange character
	if((char)PINB & (char)0x40) TXchar = 0x8F; else TXchar = 0x8E;
	Sendchar();
	TXchar = IRange + 0x41;
	Sendchar();
	// Switch 3x and 8x Amplifiers ON/OFF
	ADMUXGND;
	switch(IRange) {
	case 0: GAIN3_ON; GAIN8_ON; break;
	case 1: GAIN3_OFF; GAIN8_ON; break;
	case 2: GAIN3_ON; GAIN8_OFF; break;
	case 3: GAIN3_OFF; GAIN8_OFF; break;
	}
	// Send GND measured values
	TXchar = 0x9E;
	Send32Int();

	// Send REFERENCE measured values
	switch(IRange) {
	case 0: ADMUXR02; break;
	case 1: ADMUXR12; break;
	case 2: ADMUXR05; break;
	case 3: ADMUXR25; break;
	}
	// Average GND value
	Zero = 0;
	for(i=0;i<32;i++) Zero+=Values[i];
	Zero = Zero >> 5;
	TXchar = 0xF8;
	Send32Int();

	// Measure Value - switch input atenuator
	ADMUXCUR;
	TXchar = 0xD8;
	GetMax();

	// Test new IRange for next measure
	if((Max>350) && (IRange<3)) IRange++;
	if((Max<100) && (IRange>0)) IRange--;

	RangeFl++;
	RangeFl &= 0x03;

	// At last wait for next measure - blink LED
	if(!RangeFl) {
		LED_ON;
		for(Max=0; Max<10000; Max++) {}
		LED_OFF;
	}

  } /* loop forever */
  /* please make sure that you never leave main */
  return (0);
}

/* END */
 
